"""
Database models for persona_integration.
"""
import logging
import uuid

from django.conf import settings
from django.db import models
from model_utils.models import TimeStampedModel

from persona_integration.exceptions import InvalidStatusTransition
from persona_integration.statuses import VerificationAttemptStatus

logger = logging.getLogger(__name__)


class UserPersonaAccount(TimeStampedModel):
    """
    This model contains Persona related user information and maps to User.

    ... no_pii: This model has no PII.
    """

    # external reference to be shared between edx and Persona
    external_user_id = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, unique=True)


class VerificationAttempt(TimeStampedModel):
    """
    Information about a verification attempt (i.e. a Persona inquiry).

    .. pii: Stores full name of a user.
    .. pii_types: name
    .. pii_retirement: local_api

    """

    # Because the edx-platform verify_student.VerificationAttempt model has the same name as this model and also has a
    # foreign key to the User model, the default reverse accessors for this foreign key relationship and that foreign
    # key relationship have the same name (User.verificationattempt_set), resulting in a conflict. For this reason, we
    # set an explicit related_name value here to resolve the conflict.
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='persona_verificationattempt_set'
    )

    # The name field refers to the learner's verified name as submitted by the learner. It represents the name that the
    # learner stated will appear on their ID. The statuses approved and declined reflect whether or not Persona verified
    # that this name appears on the learner's ID.
    name = models.CharField(max_length=255)

    STATUS_CHOICES = [
        VerificationAttemptStatus.initialized,
        VerificationAttemptStatus.created,
        VerificationAttemptStatus.pending,
        VerificationAttemptStatus.expired,
        VerificationAttemptStatus.approved,
        VerificationAttemptStatus.declined,
    ]

    status = models.CharField(max_length=64, choices=[(status, status) for status in STATUS_CHOICES])

    expiration_date = models.DateTimeField(
        null=True,
        blank=True,
        db_index=True,
    )

    # The inquiry_id field refers to the ID of the inquiry in Persona. This value is used to uniquely refer to an
    # inquiry in Persona.
    inquiry_id = models.CharField(max_length=255, unique=True, blank=True, null=True)

    # The event_created_at field refers to the created_at field of the latest event received and processed via webhook
    # for this attempt. It is used to determine when events are received out of order.
    event_created_at = models.DateTimeField(null=True)

    # The platform_verification_attempt_id refers to the id of an edx-platform verify_student.VerificationAttempt model
    # instance. The reason we store this is so that we have access to the model instance id, which is needed to make
    # calls to the edx-platform update API.
    # The decision to store the id and not a traditional foreign key is documented in an ADR in this repository.
    platform_verification_attempt_id = models.PositiveIntegerField(null=True)

    def update_status(self, status):
        """
        Update the VerificationAttempt status.

        Arguments:
            * status (str): the new status

        Returns:
            * None

        Raises:
            * InvalidStatusTransition if the status transition is not valid
        """
        is_valid = VerificationAttemptStatus.is_valid_status_transition(self.status, status)

        if is_valid:
            self.status = status
            self.save()
        else:
            msg = (
                'The requested status transition is invalid. ' +
                'The current status is {from_status}, and the new status is {to_status}.'.format(
                    from_status=self.status, to_status=status
                )
            )
            raise InvalidStatusTransition(msg)



"""
Utils for test cases
"""
from unittest.mock import patch

from django.test import TestCase


class PersonaIntegrationTestCase(TestCase):
    """
    Mocks platform imports that would otherwise cause tests to fail
    """

    def setUp(self):
        """
        Setup for tests
        """
        super().setUp()

        patch.dict(
            'persona_integration.signals.handlers.LMS_STATUS_MAPPING',
            {
                'initialized': 'created',
                'created': 'created',
                'pending': 'pending',
                'declined': 'denied',
                'approved': 'approved',
            }
        ).start()
        patch('persona_integration.signals.handlers.create_verification_attempt', return_value=1).start()
        patch('persona_integration.signals.handlers.update_verification_attempt').start()



# Generated by Django 4.2.15 on 2024-10-04 16:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('persona_integration', '0006_alter_verificationattempt_user'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verificationattempt',
            name='event_created_at',
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name='verificationattempt',
            name='inquiry_id',
            field=models.CharField(blank=True, max_length=255, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='verificationattempt',
            name='status',
            field=models.CharField(choices=[('initialized', 'initialized'), ('created', 'created'), ('pending', 'pending'), ('expired', 'expired'), ('approved', 'approved'), ('declined', 'declined')], max_length=64),
        ),
    ]



"""
URLs for persona_integration.
"""
from django.urls import path

from persona_integration import views

app_name = 'persona_integration'

urlpatterns = [
    path('persona/v1/webhook', views.WebhookView.as_view(), name='webhook'),
    path('persona/v1/inquiry', views.VerificationAttemptView.as_view(), name='create_inquiry'),
]



"""
persona_integration utilities.
"""
from datetime import datetime


def parse_basic_iso_date_to_full_date(date):
    """
    Parse a date string into a timezone aware datetime object.

    This method assumes that the date string parameter follows the full ISO 8601 format.

    Below is an example string in this format.

    '2024-08-06T20:04:19.177Z'

    Note that The "Z" at the end of the date string stands for "Zulu time," which is a military and aviation term
    for Coordinated Universal Time (UTC). It indicates that the time is given in the UTC time zone, with no offset
    for local time zones.

    The datetime module supports a more limited set of the ISO 8601 standard. For example, the datetime module does not
    have native support for the "Z" suffix when parsing strings to datetime objects, so the datetime object resulting
    from parsing such a string is naive.

    For this reason, we replace the "Z" suffix with "+00:00", which represents UTC in the standard ISO format.

    Arguments:
        * date (str): the datetime string

    Returns:
        * date (datetime): the datetime object representing the date
    """
    iso_date = date.replace('Z', '+00:00')

    aware_datetime = datetime.fromisoformat(iso_date)

    return aware_datetime



# Generated by Django 4.2.14 on 2024-08-28 20:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('persona_integration', '0005_verificationattempt_platform_verification_attempt_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verificationattempt',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='persona_verificationattempt_set', to=settings.AUTH_USER_MODEL),
        ),
    ]



"""
Admin forms for persona_integration.
"""
from django.contrib import admin

from persona_integration.models import UserPersonaAccount, VerificationAttempt


@admin.register(UserPersonaAccount)
class UserPersonaAccountAdmin(admin.ModelAdmin):
    """
    Django admin form for UserPersonAccount model.
    """

    fields = ['created', 'external_user_id', 'modified', 'user',]
    list_display = ['user', 'external_user_id',]
    raw_id_fields = ['user',]
    readonly_fields = ['created', 'external_user_id', 'modified',]
    search_fields = ['external_user_id', 'user__username',]


@admin.register(VerificationAttempt)
class VerificationAttemptAdmin(admin.ModelAdmin):
    """
    Django admin form for VerificationAttempt model.
    """

    fields = [
        'created', 'event_created_at', 'expiration_date', 'inquiry_id', 'modified', 'platform_verification_attempt_id',
        'status', 'user',
    ]
    list_display = [
        'id', 'user', 'name', 'status', 'inquiry_id', 'platform_verification_attempt_id', 'created', 'modified',
        'event_created_at',
    ]
    raw_id_fields = ['user',]
    readonly_fields = ['created', 'event_created_at', 'modified', 'platform_verification_attempt_id',]
    search_fields = ['inquiry_id', 'user__username',]



"""
Factories for persona_integration tests.
"""

from datetime import timezone

import factory
from django.contrib.auth import get_user_model
from factory.django import DjangoModelFactory

from persona_integration.models import UserPersonaAccount, VerificationAttempt
from persona_integration.statuses import VerificationAttemptStatus


class UserFactory(DjangoModelFactory):
    """
    Factory to create User objects.
    """

    class Meta:
        model = get_user_model()
        django_get_or_create = (
            'email',
            'username',
        )

    _DEFAULT_PASSWORD = 'test'

    username = factory.Sequence('user{}'.format)
    email = factory.Sequence('user+test+{}@edx.org'.format)
    password = factory.PostGenerationMethodCall('set_password', _DEFAULT_PASSWORD)
    is_superuser = False
    is_staff = False


class UserPersonaAccountFactory(DjangoModelFactory):
    """
    Factory to crate UserPersonaAccount objects.
    """

    class Meta:
        model = UserPersonaAccount

    user = factory.SubFactory(UserFactory)
    external_user_id = factory.Faker('uuid4')


class VerificationAttemptFactory(DjangoModelFactory):
    """
    Factory to create VerificationAttempt objects.
    """
    class Meta:
        model = VerificationAttempt

    user = factory.SubFactory(UserFactory)
    name = factory.Faker('name')
    status = VerificationAttemptStatus.created
    expiration_date = None
    inquiry_id = factory.Sequence('inquiry-id-{}'.format)
    event_created_at = factory.Faker('date_time', tzinfo=timezone.utc)



"""
Status enums for persona_integration.
"""


class VerificationAttemptStatus:
    """This class describes valid statuses for a verification attempt to be in."""

    # This is the initial state of a verification attempt, after a learner has requested to complete IDV.
    initialized = "initialized"

    # A webhook was received that confirms that an inquiry has been created in Persona.
    created = "created"

    # A verification attempt is pending when it has been started but has not yet been completed.
    pending = "pending"

    # A verification attempt is expired when it is has not been completed in a pre-configured amount of time.
    expired = "expired"

    # A verification attempt is approved when it has been manually approved or approved via a Persona workflow.
    approved = "approved"

    # A verification attempt is declined when it has been manually declined or declined via a Persona workflow.
    declined = "declined"

    all_statuses = [initialized, created, pending, expired, approved, declined]
    terminal_statuses = [expired, approved, declined]

    @classmethod
    def is_valid_status(cls, status):
        """
        Return whether a verification attempt status is valid.

        Arguments:
            * status (str): the attempt status

        Returns:
            * True if the status is valid; else, False
        """
        return status in cls.all_statuses

    @classmethod
    def is_terminal_status_transition(cls, from_status, to_status):
        """
        Return whether the status transition is a terminal status transition.

        A terminal status transition is a transition between two terminal states. This implies that an override was
        performed via the Persona dashboard, as this kind of status transition should not occur during the usual process
        of IDV.

        Arguments:
            * from_status (str): the current verification attempt status
            * to_status (str): the new verification attempt status

        Returns:
            * True if the status transition is between two terminal states; else, False
        """
        terminal_statuses = set(cls.terminal_statuses)
        return from_status in terminal_statuses and to_status in terminal_statuses

    @classmethod
    def is_valid_status_transition(cls, from_status, to_status):
        """
        Return whether a verification attempt status transition is valid.

        Arguments:
            * from_status (str): the current verification attempt status
            * to_status (str): the new verification attempt status

        Returns:
            * True if the status transition is valid; else, False
        """
        terminal_statuses = set(cls.terminal_statuses)

        if from_status == cls.initialized:
            return to_status in set([cls.created, cls.pending, cls.expired, cls.approved, cls.declined])
        if from_status == cls.created:
            return to_status in set([cls.pending, cls.expired, cls.approved, cls.declined])
        elif from_status == cls.pending:
            return to_status in set([cls.expired, cls.approved, cls.declined])
        elif from_status in terminal_statuses:
            # We allow terminal state transitions in order to allow for overrides via the Persona dashboard.
            return to_status in terminal_statuses
        else:
            # This code path should never be reached, because the above conditions list all possible statuses.
            return False



"""
Views for persona_integration.
"""
import hmac
import logging
from collections import namedtuple

from django.conf import settings
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import status
from rest_framework.authentication import SessionAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from persona_integration.api import create_inquiry, get_user_by_external_id, update_verification_attempt
from persona_integration.exceptions import InvalidStatusTransition, PersonaCannotRegisterAttempt
from persona_integration.utils import parse_basic_iso_date_to_full_date

log = logging.getLogger(__name__)


class VerificationAttemptView(APIView):
    """
    When called, create an inquiry for a user who needs to undergo the ID Verification process.

    Persona strongly recommended that we preemptively create an inquiry on their backend before
    a learner accesses the Person application to perform IDV via our own backend call,
    because calls to their inquiry endpoint cannot come from the browser.

    Accepts:
        * [POST]

    This class only supports POST requests.
    """

    authentication_classes = (SessionAuthentication, JwtAuthentication)
    permission_classes = (IsAuthenticated,)

    def post(self, request):
        """
        Receive and pass along requests to create an inquiry in Persona.

        Paths:
        * persona/create_inquiry

        Returns:
        * 201: The request to create an inquiry in Persona was successful.
        (NOTE: This backend will still receive a request in the WebhookView to confirm that an inquiry was created.)
        * 400: The request data is invalid.
        * 500: Failed to create Persona Inquiry from this backend.
        """
        user = request.user
        verified_name = request.data.get('verified_name')
        idempotency_key = request.data.get('idempotency_key')

        # If there is a bad request from the UI to this backend, return 400:
        if None in [verified_name]:
            return Response(
                status=status.HTTP_400_BAD_REQUEST,
                data={'detail': (
                    'Bad request to persona-integration. verified_name is a required field.'
                )}
            )

        try:
            _, inquiry_id = create_inquiry(user, verified_name, idempotency_key)
        # If the request from this backend to Persona fails, return 500
        except PersonaCannotRegisterAttempt:
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return Response(status=status.HTTP_201_CREATED, data={'inquiry_id': inquiry_id})


@method_decorator(csrf_exempt, name='dispatch')
class WebhookView(APIView):
    """
    Receive and process webhook requests from Persona, corresponding to lifecycle events.

    Currently, this view only processes inquiry related events, as configured in the Persona dashboard.

    Accepts:
        * [POST]

    This class only supports POST requests.
    """

    Event = namedtuple('Event', [
        'webhook_type',
        'event_id',
        'payload_type',
        'inquiry_id',
        'inquiry_status',
        'created_at',
        'reference_id',
    ])

    def validate_request_signature(self, request, webhook_secret):
        """
        Return whether the Persona-Signature request header is valid.

        This function supports keys in the process of rotation.

        The format of the Persona-Signature header is as follows.

        t=<timestamp>, v1=<signature>

        If the header has two signatures, which will happen if the keys are in the process of rotatin, the format of the
        Persona-Signature header is as follows.

        t=<timestamp>, v1=<signature_new> v1=<signature_old>

        Arguments:
            * request (Request): request object
            * webhook_secret (str): Persona webhook secret used to generate the HMAC

        Returns:
            * True if the signature is valid; else, False
        """
        timestamp, signatures = list(request.headers['Persona-Signature'].split(','))

        timestamp = timestamp.split('t=')[1]
        digest = hmac.new(
            webhook_secret.encode(), (timestamp + '.' + request.body.decode('utf-8')).encode(), 'sha256'
        ).hexdigest()

        if len(signatures.split(' ')) > 1:
            signature_new, signature_old = [value.split('v1=')[1] for value in signatures.split(' ')]
            if hmac.compare_digest(signature_new, digest) or hmac.compare_digest(signature_old, digest):
                return True
        else:
            signature = signatures.split('v1=')[1]
            signature = request.headers['Persona-Signature'].split(',')[1].split('=')[1]
            if hmac.compare_digest(signature, digest):
                return True

        return False

    def extract_event_data(self, request):
        """
        Extract relevant data from the event.

        Arguments:
            * request (Request): request object

        Returns:
            * an instance of the Event namedtuple; False if event data is missing or cannot be extracted
        """
        try:
            data = request.data['data']
            webhook_type = data['type']
            event_id = data['id']
            payload_type = data['attributes']['payload']['data']['type']

            inquiry_id = data['attributes']['payload']['data']['id']
            inquiry_status = data['attributes']['payload']['data']['attributes']['status']
            reference_id = data['attributes']['payload']['data']['attributes']['reference_id']

            created_at = parse_basic_iso_date_to_full_date(data['attributes']['created_at'])

            return WebhookView.Event(
                webhook_type=webhook_type,
                event_id=event_id,
                payload_type=payload_type,
                inquiry_id=inquiry_id,
                inquiry_status=inquiry_status,
                created_at=created_at,
                reference_id=reference_id,
            )
        except KeyError:
            return False

    def post(self, request):
        """
        Receive and process webhook requests from Persona, corresponding to lifecycle events.

        Currently, this view only processes inquiry related events, as configured in the Persona dashboard.

        A response with a 200 status code will be returned for all valid requests (i.e. veriable signature, properly
        formed request) to ensure that Persona does not re-try the request, but only inquiry related events will be
        processed.

        Note that a signature in the Persona-Signature header is required.

        Paths:
        * persona/webhook

        Returns:
        * 200: The request was successfully handled.
        * 400: The request data is malformed or invalid.
        * 404: The request could not be authenticated.
        * 500: The server is misconfigured (e.g. the webhook secret configuration is missing).
        """
        webhook_secret = settings.VERIFY_STUDENT.get('PERSONA', {}).get('WEBHOOK_SECRET')

        if not webhook_secret:
            log.error('Persona webhook secret is missing.')
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        is_valid = self.validate_request_signature(request, webhook_secret)

        if not is_valid:
            log.error('Persona webhook signature is not valid.')
            return Response(status=status.HTTP_404_NOT_FOUND)

        event = self.extract_event_data(request)

        if not event:
            log_msg = 'Persona event is malformed and is missing required data.'

            event_id = request.data.get('id')
            if event_id:
                log_msg += 'The event ID is %s.'
                log.error(log_msg, event_id)
            else:
                log.error(log_msg)

            return Response(status=status.HTTP_400_BAD_REQUEST)

        # If the webhook data does not represent an event or is not related to an inquiry, ignore the request
        # and return a response with a 200 status to acknowledge the webhook.
        if event.webhook_type != 'event' or event.payload_type != 'inquiry':
            return Response(status=status.HTTP_200_OK)

        # Log the event for auditability. Although we have set up inquiry blocklists to filter out any extraneous PII,
        # we still avoid logging the entire event object to ensure that we do not accidentally log any additional PII.
        log_msg = (
            'Received event from Persona. '
            'event ID: %s, created at: %s, inquiry ID: %s, reference_id: %s, inquiry status: %s',
        )
        log.info(
            log_msg,
            event.event_id,
            event.created_at,
            event.inquiry_id,
            event.reference_id,
            event.inquiry_status,
        )

        user = get_user_by_external_id(event.reference_id)

        if not user:
            log_msg = (
                'The reference_id in the Persona event is not associated with a user. '
                'The event ID is %s.'
            )
            log.error(log_msg, event.event_id)
            return Response(status=status.HTTP_400_BAD_REQUEST)

        try:
            update_verification_attempt(
                event.inquiry_status,
                event.inquiry_id,
                event.created_at,
            )
        except InvalidStatusTransition as e:
            # Theoretically, this should not occur, because out of order events are disregarded, and the verification
            # attempt lifecycle is unidirectional. In this case, we return a 200 to Persona to acknowledge the webhook
            # and log the case.
            log_msg = (
                'The status in the Persona event represents an invalid status transition. This should not occur. '
                'The event ID is %s. '
                'Exception: %s.'
            )
            log.warning(log_msg, event.event_id, e)

        return Response(status=status.HTTP_200_OK)



"""
Python API for persona_integration.
"""
import json
import logging
from datetime import timedelta

import requests
from django.conf import settings

from persona_integration.exceptions import InvalidStatusTransition, PersonaCannotRegisterAttempt
from persona_integration.models import UserPersonaAccount, VerificationAttempt
from persona_integration.statuses import VerificationAttemptStatus

log = logging.getLogger(__name__)


def create_inquiry(user, verified_name, idempotency_key=None):
    """
    Send a request to Persona to create an Inquiry with the user's information.

    Arguments:
        * user (User): the user (usually a learner) who will be performing the inquiry
        * verified_name (string): the verified name of the user
        * idempotency_key (string): the key used to make sure that inquiries are idempotent,
          i.e. ensure that Persona ignores requests with the same idempotency key so repeat
          requests (i.e. retries) do not create a new inquiry for each indentical requests
          meant to represent a single inquiry. It is optional and defaults to None.

    Returns:
        * status (string): the response code from Persona

    Note that this api function may also raise the PersonaCannotRegisterAttempt exception
    if the request to Persona fails.
    """
    inquiry_template_id = settings.VERIFY_STUDENT['PERSONA']['INQUIRY_TEMPLATE_ID']
    api_key = settings.VERIFY_STUDENT['PERSONA']['API_KEY']

    user_persona_account, _ = UserPersonaAccount.objects.get_or_create(user=user)
    account_reference_id = user_persona_account.external_user_id

    headers = {
        'accept': 'application/json',
        'Persona-Version': '2023-01-05',
        'Key-Inflection': 'snake',  # Snake case in Python, then change to camel case in frontend.
        'authorization': f'Bearer {api_key}',
    }

    # The idempotency key allows us to retry this request in the case of a network failure. It is optional.
    if idempotency_key:
        headers['Idempotency-Key'] = idempotency_key

    payload = {
        'data': {
            'attributes': {
                'inquiry_template_id': inquiry_template_id,
                'fields': {
                    'name_first': verified_name,
                },
            },
        },
        'meta': {
            'auto-create-account-reference-id': str(account_reference_id),
        }
    }

    api_root = settings.VERIFY_STUDENT['PERSONA']['API_ROOT']
    url = f'{api_root}/inquiries'

    # Set up a VerificationAttempt, but don't save it unless the request succeeds & if we're not resuming an inquiry.
    new_verification_attempt = VerificationAttempt(
        user=user,
        name=verified_name,
        status=VerificationAttemptStatus.initialized,
    )

    # To keep things simple, this application does not handle retrying this request if it fails.
    # The client may handle retrying this request via the VerificationAttemptView if it fails.
    try:
        response = requests.post(url, headers=headers, json=payload, timeout=10)

        # If Persona returns an non-success status code, that means we coded something wrong in this backend.
        # In which case, we should log the response and raise an exception.
        if response.status_code >= 300:
            log.error(
                'Persona failed to create an inquiry with the following response: %(response_msg)s '
                'with status code: %(http_status)s',
                {
                    'http_status': response.status_code,
                    'response_msg': json.dumps(response.json()),
                }
            )
            raise PersonaCannotRegisterAttempt

        inquiry_id = response.json().get('data', {}).get('id')

    # If the request to Persona doesn't return anything and fails for another reason, something is likely wrong
    # on Persona's end. In this case, we should log a separate error.This captures several errors that could be
    # raised by the `requests` package: https://3.python-requests.org/user/quickstart/#errors-and-exceptions
    except Exception as error_message:
        log.error(
            'Failed to connect to Persona with error message: %(error_message)s',
            {'error_message': str(error_message)}
        )
        raise PersonaCannotRegisterAttempt from error_message

    log.info(
        'Request to Persona to create an inquiry (IDV attempt) with inquiry_id=%(inquiry_id)s '
        'for user with user_id=%(user_id)s was successful with response status=%(response_status)s)',
        {
            # NOTE: The response.data object as a whole contains PII, so only ever send back the id
            'inquiry_id': inquiry_id,
            'user_id': user.id,
            'response_status': response.status_code,
        }
    )

    # If a VerificationAttempt with the returned inquiry_id already exists, that means this endpoint was called
    # more than once with the same idempotency key. In this case, we simply return the inquiry_id. Otherwise,
    # if no such attempt exists, then that means we sent Persona a new, freshly generated idempotency key,
    # or the idempotency key was the same but was more than an hour old and was garbage collected by Persona.
    try:
        VerificationAttempt.objects.get(inquiry_id=inquiry_id)
    except VerificationAttempt.DoesNotExist:
        new_verification_attempt.inquiry_id = inquiry_id
        new_verification_attempt.save()

    return response.status_code, inquiry_id


def get_expiration_date(date):
    """
    Return the expiration date of a verification attempt given a date.

    Arguments:
        * date (datetime): the date from which an expiration date is calculated

    Returns:
        * expiration date (datetime)
    """
    return date + timedelta(days=settings.VERIFY_STUDENT["DAYS_GOOD_FOR"])


def get_user_by_external_id(external_user_id):
    """
    Get a user associated with an external_user_id.

    Arguments:
        * external_user_id (UUID): the external user ID that is sent to Persona and uniquely identifies the user

    Returns:
        * user (User) if one exists; else, None
    """
    try:
        return UserPersonaAccount.objects.get(external_user_id=external_user_id).user
    except UserPersonaAccount.DoesNotExist:
        return None


def update_verification_attempt(status, inquiry_id, event_created_at):
    """
    Update a Verification Attempt.

    Update the status of the latest VerificationAttempt associated with an inquiry_id, user, and name.
    If the attempt is moving to the "created" status, also update event_created_at.

    Arguments:
        * status (str): the status of the inquiry
        * inquiry_id (str): the ID of the Persona inquiry,
        * event_created_at (datetime): the created_at field of the event

    Returns:
        * None
    """
    if not VerificationAttemptStatus.is_valid_status(status):
        msg = (
            'The requested status {status} is invalid. '.format(
                status=status
            )
        )
        raise InvalidStatusTransition(msg)

    verification_attempt = VerificationAttempt.objects.get(inquiry_id=inquiry_id)

    # Do not apply the update if the event is a duplicate or if the associated event was received out of order.
    # Otherwise, perform the update.
    event_is_duplicate = verification_attempt.status == status

    # NOTE: The VerificationAttempt's event_created_at might be None if it was just initialized,
    # so this boolean prevents us from comparing None to a valid datetime, which would throw an error.
    event_received_out_of_order = (verification_attempt.event_created_at is not None
                                   and event_created_at < verification_attempt.event_created_at)
    if event_is_duplicate or event_received_out_of_order:
        return

    # Calculate the expiration date using the created_at timestamp, which is the time at which the approved event
    # was created. This ties the expiration to the time the approval event is created and not to the time that the
    # event is received.
    if status == VerificationAttemptStatus.approved:
        expiration_date = get_expiration_date(event_created_at)
        verification_attempt.expiration_date = expiration_date

    if VerificationAttemptStatus.is_terminal_status_transition(verification_attempt.status, status):
        log.info(
            'The status transition for inquiry (IDV attempt) with inquiry_id=%(inquiry_id)s for user with '
            'user_id=%(user_id)s is a terminal status transition. The transition is %(from_status)s to %(to_status)s.',
            {
                'inquiry_id': inquiry_id,
                'user_id': verification_attempt.user.id,
            }
        )

    verification_attempt.event_created_at = event_created_at
    verification_attempt.update_status(status)
    verification_attempt.save()



"""
Signal handlers for the persona-integration plugin.
"""
from django.db.models.signals import post_save
from django.dispatch import Signal, receiver

from persona_integration.models import VerificationAttempt
from persona_integration.statuses import VerificationAttemptStatus

try:
    from lms.djangoapps.verify_student.api import create_verification_attempt, update_verification_attempt
    from lms.djangoapps.verify_student.statuses import VerificationAttemptStatus as PlatformVerificationAttemptStatus
except ImportError:
    create_verification_attempt = None
    update_verification_attempt = None
    PlatformVerificationAttemptStatus = None

try:
    from openedx.core.djangoapps.user_api.accounts.signals import USER_RETIRE_LMS_MISC
except ImportError:
    # An ImportError should only be raised in tests, where the code is not running as an installation of
    # edx-platform. In this case, the import should default to a generic Signal.
    USER_RETIRE_LMS_MISC = Signal()

LMS_STATUS_MAPPING = {
    # This may seem redundant, but we must synchronize the initialized statue to the LMS because of the branching logic
    # in the on_verification_attempt_change signal handler, which determines which platform API method to call based on
    # whether the instance of the VerificationAttempt was created. Because a VerificationAttempt is always created
    # with the initialized status, in order for the branch that calls create_verification_attempt to be invoked, we must
    # synchronize the initialized status.
    VerificationAttemptStatus.initialized: (
        PlatformVerificationAttemptStatus and PlatformVerificationAttemptStatus.CREATED
    ),
    VerificationAttemptStatus.created: (
        PlatformVerificationAttemptStatus and PlatformVerificationAttemptStatus.CREATED
    ),
    VerificationAttemptStatus.pending: (
        PlatformVerificationAttemptStatus and PlatformVerificationAttemptStatus.PENDING
    ),
    VerificationAttemptStatus.approved: (
        PlatformVerificationAttemptStatus and PlatformVerificationAttemptStatus.APPROVED
    ),
    VerificationAttemptStatus.declined: (
        PlatformVerificationAttemptStatus and PlatformVerificationAttemptStatus.DENIED
    )
}


@receiver(USER_RETIRE_LMS_MISC)
def on_user_retirement(sender, **kwargs):  # pylint: disable=unused-argument
    """
    Receives a user retirement signal and retires user data
    """
    user = kwargs.get('user')
    VerificationAttempt.objects.filter(user_id=user.id).delete()


@receiver(post_save, sender=VerificationAttempt)
def on_verification_attempt_change(sender, instance, created, **kwargs):  # pylint: disable=unused-argument
    """
    Receives a verification attempt change signal and updates verification attempt status.
    """
    # NOTE: We do not update on "pending" because "pending" for Persona means that IDV in progress, which does
    # not correspond with the "pending" status in the LMS, which indicates that the user has submitted an IDV attempt
    # and blocks the user from submitting another verified name in their account settings. This prevents a user from
    # being blocked from starting a new inquiry while one is still in progress/not yet submitted.
    statuses_to_update = [
        VerificationAttemptStatus.created,
        VerificationAttemptStatus.approved,
        VerificationAttemptStatus.declined
    ]

    if instance.status in LMS_STATUS_MAPPING:
        if created:
            attempt_id = create_verification_attempt(
                user=instance.user,
                name=instance.name,
                status=LMS_STATUS_MAPPING[instance.status],
                expiration_datetime=instance.expiration_date
            )
            instance.platform_verification_attempt_id = attempt_id
            instance.save()
        else:
            if instance.status in statuses_to_update:
                update_verification_attempt(
                    attempt_id=instance.platform_verification_attempt_id,
                    name=instance.name,
                    status=LMS_STATUS_MAPPING[instance.status],
                    expiration_datetime=instance.expiration_date
                )



# Generated by Django 4.2.14 on 2024-08-07 18:43

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('persona_integration', '0003_verificationattempt_event_created_at_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userpersonaaccount',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, unique=True),
        ),
    ]



"""
Exceptions for persona_integration.
"""


class InvalidStatusTransition(Exception):
    """
    This exception is raised if a status transition is not valid.
    """


class PersonaCannotRegisterAttempt(Exception):
    """
    Raised when a Persona cannot register an attempt for any reason.
    """



# Generated by Django 4.2.14 on 2024-08-28 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('persona_integration', '0004_alter_userpersonaaccount_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='verificationattempt',
            name='platform_verification_attempt_id',
            field=models.PositiveIntegerField(null=True),
        ),
    ]



"""
persona_integration Django application initialization.
"""

from django.apps import AppConfig


class PersonaIntegrationConfig(AppConfig):
    """
    Configuration for the persona_integration Django application.
    """

    name = 'persona_integration'

    plugin_app = {
        'url_config': {
            'lms.djangoapp': {
                'namespace': 'persona_integration',
                'regex': '^api/',
                'relative_path': 'urls',
            },
        }
    }

    def ready(self):
        """
        Connect signal handlers.
        """
        from persona_integration.signals import handlers  # pylint: disable=import-outside-toplevel,unused-import



# Generated by Django 4.2.14 on 2024-07-31 10:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('persona_integration', '0002_verificationattempt'),
    ]

    operations = [
        migrations.AddField(
            model_name='verificationattempt',
            name='event_created_at',
            field=models.DateTimeField(default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='verificationattempt',
            name='inquiry_id',
            field=models.CharField(default=None, max_length=255, unique=True),
            preserve_default=False,
        ),
    ]
